// HID serial transfer, suitable for low-speed devices.
// Unluckily, no terminal emulator supports HID.
// Here is no OUT endpoint/report so receiving data from host
// (i.e. keyboard, cursor position report) is not implemented here.
// Maximum transfer rate: 32 kByte/s

#include "usb.h"
#include <avr/io.h>
#include <avr/pgmspace.h>
#include <avr/interrupt.h>

// machine dependendent actions
inline void HW_CONFIG() {}
inline void PLL_CONFIG() {PLLCSR = 1<<PLLE | 1<<PLLP0;}
inline void USB_CONFIG() {USBCON = 1<<USBE;}
inline void USB_FREEZE() {USBCON = 1<<USBE | 1<<FRZCLK;}

// Vier Schalter
#define HID 0	// HID benutzen oder nicht?
#define WUB 1	// WebUSB/WinUSB benutzen oder nicht? (Geht noch nicht solo!!)
#define CDC 0	// CDC benutzen oder nicht?
#define IAD 0//(HID|WUB)	// Interface Association Descriptor benutzen oder nicht? (Nur in Verbindung mit CDC)

/* Erklärung:
Das praktische Ziel ist ein USB-Gerät, welches treiberlos als HID-Gerät
(ab Windows 98), als WebUSB-Gerät (ab Windows 8.1) und als serielle
Schnittstelle (im USB-Sprech „CDC“ = Communication Device Class,
genauer „CDC-ACM“) funktioniert.

Dabei macht USB-CDC die größten Probleme:
- Solo funktioniert USB-CDC bereits ab Windows 98, erfordert jedoch
  stets eine INF-Datei sowie bis hin zu Windows 2000 einen Treiber (.sys).
  Man kennt dies von älteren USB-RS232-Adaptern.
  * Für Windows 98, Me und 2k musste der Adapterhersteller selber einen
    Treiber schreiben oder einen von Microsoft kaufen
  * Im Kaufvertrag stand ein Passus, der es verbietet, den Treiber
    generisch mit /allen/ USB-CDC-Adaptern (USB\Class_02)
    zu verbinden; die Verbindung /musste/ via eineindeutiger
    Vendor-ID + Product-ID erfolgen. Das maximierte Microsofts Absatz.
  * Ab Windows XP war usbser.sys als Treiber bereits im System vorhanden.
    Hersteller von USB-Seriell-Adaptern /mussten/ (laut Vertrag)
    ihren eigenen Treiber zurückziehen sowie die INF-Datei für
    64 Bit anpassen: Gesonderte Sections sind da erforderlich.
  * (Nicht nur) ich habe schließlich irgendeine dieser INF-Dateien
    hergenommen und generisch gemacht. Damit lassen sich USB-Seriell-Adapter
    nahezu jedes Herstellers sogar unter Windows 98 ohne lästiges
    Installationsgewurstel verwenden.
  * COM-Portnummern wechseln generell mit der Buchse.
    Hat das USB-Gerät eine Seriennummer (Device Descriptor Offset 16),
    merkt sich Windows (ab 2k) die Beziehung zwischen Nummer und jener
    Zeichenkette. Daher kommt es, dass einige CDC-Geräte ihre
    COM-Portnummer beim Umstöpseln behalten, die meisten jedoch nicht.

- In Kombination mit anderen USB-Funkionen ist USB-CDC nach USB-Standard
  anfangs nicht so recht vorgesehen. Denn CDC benötigt /zwei/ USB-
  Interfaces und kann so nicht Teil eines Muktifunktions-USB-Gerätes
  (USB\Class_00) sein: Windows lädt für die Geräteklasse 00 seinen
  Multifunktionstreiber (usbccgp.sys), der wiederum einen USB-Upstream-Port
  emuliert und für jedes enthaltene USB-Interface einzeln den passenden
  Treiber lädt. Daraufhin wird zwar usbser.sys für das entsprechende
  Control-Interface gefunden und geladen; für das Data-Interface wird
  jedoch nur ein Stub-Treiber geladen, der den sattsam bekannten
  Gerätetreiber-Installieren-Dialog hervorbringt.
  Und usbser.sys verweigert die Arbeit, weil es kein Data-Interface findet
  oder die Arbeit unter der Fuchtel von usbccgp.sys verweigert.
  Die (angeblich) richtige Lösung hierfür sei eine maßgeschneiderte
  INF-Datei, da habe ich aber nichts greifbares gefunden.

- Mit Windows 8.1, bzw. um diese Zeit wurde der USB-Standard 2.0 um einen
  „Interface Association Descriptor“ (IAD) erweitert.
  Damit soll genau dieses Problem angegangen werden:
  Der IAD gruppiert mehrere Interfaces zu dem Zweck, dass usbccgp.sys
  die Interfaces nicht zerteilt sondern als Bundle behält und dieses
  dem nachfolgenden Treiber (hier: usbser.sys) übergibt.
  Klar, usbccgp.sys hätte auch selber so schlau sein können,
  die Klasse 02 (CDC sowie CDC-Control) und die Klasse 0A (CDC-Data)
  zusammen lassen zu können. Dafür gibt es sogar der Klasse 02
  untergeordnete Deskriptoren. Hat aber Microsoft wegignoriert und lieber
  im USB-Konsortium insistiert: Will /übergeordneten/ Desktiptor.
  Bei Linux ist man genau jenen Weg gegangen, und auch für eingebettete
  Systeme ist das der gängige Weg.
  Der Pferdefuß ist nun, dass der IAD erfordert, dass das USB-Gerät
  sich mit USB-Standard 2.0 (statt wie sonst 1.1) meldet
  und dass die USB-Geräteklasse nicht mehr 00 sondern EF sein muss.
  Damit erscheint bei Mutifunktionsgeräten mit HID und CDC und IAD
  alles problemlos unter Windows 8.1+, jedoch funktionert darunter
  /gar nichts/ mehr: Denn auch das HID-Interface verschwindet,
  da bspw. Windows98 mit der neuen Geräteklasse EF nichts anfangen kann.
  Kann man zwar beheben, indem man eine INF-Datei beilegt, die
  die USB-Geräteklasse EF mit usbccgp.sys verbindet, aber das verfehlt
  erstmalig das Ziel, mit HID einen treiberlosen Zugang für /jede/
  Windows-Version bereit zu stellen.

Die Moral von der Geschicht: Entweder man benutze CDC solo, oder man
verzichte darauf. Jede Art von Kombination schränkt die Portabilität ein.
Wenn schon kombiniert, dann allenfalls als Debug-Interface, nicht für das
Fertigprodukt.

Für Mikrocontroller ohne USB-Hardware, typischwerweise mit obdev.at V-USB,
ist CDC nur bis Windows XP verwendbar.
Höhere Windows-Versionen verbieten das, was auch per USB-Standard verboten
ist, zumal CDC auf Low-Speed-USB eine hohe Belastung für den Bus
und den Mikrocontroller darstellt: Jener wird nahezu lückenlos mit
Datenlese-Tokens bombardiert, die er mit NAK beantworten muss.
(Trifft auf den hier verwendeten AT90USB82 und Nachfolger nicht zu.)

Was wäre die /richtige/ Lösung gewesen?
Richtig gewesen wäre eine Neustrukturierung der USB-CDC-Geräteklasse
sowie aller anderen USB-Klassen, die mehrere Interfaces belegen
(USB-Audio, USB-Video), zu einer 1:1-Beziehung, also Integration der
Dateninterfaces in das jeweilige Steuer-Interface.
Wurde bis heute nicht darüber nachgedacht.
Aber USB ist wie Unicode ein Politikum: Es wird nur 'rangeflickt, nichts
aufgeräumt. Nur dass das bei Unicode von vornherein angesagtes Ziel ist.
Beim USB-Standard habe ich hingegen keinen derartigen Passus gelesen.

Die /zweite richtige Lösung/ wäre ein HID-Substandard für eine
Terminal-Emulation.
Und ein PuTTY, welches derartige HID-Geräte ansprechen kann.
Zwar mit vermindertem Durchsatz gegenüber CDC, aber ohne den ganzen Ärger:
* Zwang zu USB Full-Speed
* Wechselnde COM-Portnummern
* Oben angeführte Kompatibilitätsprobleme in Multifunktionsgeräten


*WebUSB* ist ebenfalls ein heißes Eisen, verträgt sich allerdings
von vornherein mit einem benachbarten HID-Interface und stellt so
maximale Portabilität sicher.
Ein WebUSB-Interface ist der Schlüssel für den Zugriff auf ein USB-Gerät
mittels Smartphone oder Smart-TV. Das kann bei Bastelprojekten das lästige
Design eines Benutzerinterfaces in Hardware ersparen oder reduzieren.
Als Beispiel sei ein Rasenmähroboter genannt, der zur Lokalbedienung
nur 3 Knöpfe (Stop, Home, Work) und wenige LED
(Homing, Working, Collision+Lift Detect, Battery Indicator) benötigt.
Alles andere (Uhrzeit, Kalender, Passwort, dGPS-Setup wenn vorhanden)
dann mittels Smartphone.

Zu beachten ist, dass Firefox seine Zusammenarbeit mit WebUSB bis heute
verweigert. Alle anderen Browser können das mittlerweile.
Der erste war Google Chrome, dicht gefolgt von Microsoft Edge.
WebUSB funktioniert etwa ab Android 7 auf Galaxy A5 (2016).


 *  Endpoint Buffer Configuration
Der AT90USB162 hat nur wenig Endpoint-RAM (176 Bytes)
und nur wenig (fünf) Endpoints, davon nur zwei mit Doppelpufferung.
Trotzdem reicht das gut für ein HID+CDC-Kombidevice oder ein Audio-Device.
Als sinnvolle Endpoint-Größe erscheint:
Anwendung			EP0	EP1	EP2	EP3	EP4
CDC, symmetrisch		32	-	16	32×2	32×2
CDC, asymmetrisch		8	-	8	64×2	16×2
HID+CDC, symmetrisch		16	16	16	32×2	32×2
HID+CDC, asymmetrisch		32	32	16	32×2	16×2
HID+CDC, stark asymmetrisch	8	16	8	64×2	8×2
WebUSB+CDC (Oszilloskop)	8	16	8	64×2	8×2
WebUSB+HID+CDC (Oszilloskop)	Nicht empfehlenswert, da fehlt ein Endpoint
HiFi-Stereo-Mikrofon		32	16	-	64×2?	-
HiFi-Lautsprecher/Kopfhörer	32	16	-	-	64×2?
Headset, Telefon		32	16	-	64	64
Kamera, Massenspeicher		Nicht mit USB Full-Speed empfehlenswert
 **************************************************************************/

// you might want to change the buffer size, up to 64 bytes.
// The host reserves your bandwidth because this is an interrupt
// endpoint, so it won't be available to other interrupt or isync
// endpoints in other devices on the bus.
enum{
 EP0SSH	= 5,	// EP0: Control 32 Byte
 EP1SSH = 5,	// EP1: HID-Interrupt 32 Byte
 EP2SSH = 4,	// EP2: CDC-Status 16 Byte
 EP3SSH	= 5,	// EP3: CDC-Daten µC->PC 32×2 Byte
 EP4SSH = 4,	// EP4: CDC-Daten PC->µC 16×2 Byte
};

/**************************************************************************
 *  Descriptor Data
 **************************************************************************/

// Descriptors are the data that your computer reads when it auto-detects
// this USB device (called "enumeration" in USB lingo).  The most commonly
// changed items are editable at the top of this file.  Changing things
// in here should only be done by those who've read chapter 9 of the USB
// spec and relevant portions of any USB class specifications!

#define W(x) (byte)(x),(byte)((x)>>8)

static const byte PROGMEM DeviceDesc[] = {
 18,		// bLength
 1,		// bDescriptorType
#if WUB
 W(0x0210),	// bcdUSB		2.1
#elif CDC*IAD
 W(0x0200),	// bcdUSB		2.0
#else
 W(0x0110),	// bcdUSB		1.1
#endif
#if CDC		// Class,SubClass,Protocol 
# if IAD
 239,2,1,	// misc. device		Not understood by Windows <=XP!
# else
 2,2,1,		// communication device
# endif
//#elif HID*(1-WUB)
// 3,0,0,
//#elif (1-HID)*WUB
// 0xFF,0,0,
#else
 0,0,0,		// see interface descriptor(s)
#endif
 1<<EP0SSH,	// bMaxPacketSize0
 W(5824),	// idVendor		Voti
#if HID*CDC
 W(10200),	// idProduct		obdev: libusb, here multifunction
#elif CDC
 W(10205),	// idProduct		obdev: CDC-ACM
#else
 W(0x27D9),	// idProduct		obdev: HID no keyboard, mouse, joystick
#endif
 W(0x0100),	// bcdDevice		1.0
 1,		// iManufacturer
 2,		// iProduct
 3,		// iSerialNumber	So Windows COM port won't move
 1,		// bNumConfig
};

#if HID
static const byte PROGMEM HidRepDesc[] = {
 0x06, W(0xFF31),	// Usage Page 0xFF31 (vendor defined)
 0x09, 0x74,		// Usage 0x74
 0xA1, 0x53,		// Collection 0x53
  0x75, 8,		// report size = 8 bits
  0x15, 0,		// logical minimum = 0
  0x26, W(255),		// logical maximum = 255
  0x95, 1<<EP1SSH,	// report count = 32
  0x09, 0x75,		// usage
  0x81, 0x02,		// Input (array)
 0xC0			// end collection
};
# define HID_DESC_OFFSET  (9+9)
#endif

static const byte PROGMEM ConfigDesc[9+(9+9+7)*HID+(9+2*7*!CDC)*WUB
  +(8*IAD+9+5+4+5+5+7+9+7+7)*CDC] = {
// configuration descriptor, USB spec 9.6.3, page 264-266, Table 9-10
 9, 			// bLength
 2,			// bDescriptorType	configuration
 W(sizeof ConfigDesc),	// wTotalLength
 HID+WUB+2*CDC,		// bNumInterfaces	HID + CDC Control + CDC Data
 1,			// bConfigurationValue
 0,			// iConfiguration	no string descriptor
 0xC0,			// bmAttributes		Bit6: SelfPower
 50,			// bMaxPower		100 mA
#if HID
// interface descriptor, USB spec 9.6.5, page 267-269, Table 9-12
 9,			// bLength
 4,			// bDescriptorType
 0,			// bInterfaceNumber
 0,			// bAlternateSetting
 1,			// bNumEndpoints
 3,			// bIfaceClass		HID
 0,			// bIfaceSubClass	no boot support
 0,			// bIfaceProtocol
 4*(WUB|CDC),		// iInterface
// HID interface descriptor, HID 1.11 spec, section 6.2.1
 9,			// bLength
 0x21,			// bDescriptorType
 W(0x0110),		// bcdHID
 0,			// bCountryCode
 1,			// bNumDescriptors
 0x22,			// bDescriptorType	Report
 W(sizeof(HidRepDesc)),	// wDescriptorLength
// endpoint descriptor, USB spec 9.6.6, page 269-271, Table 9-13
 7,			// bLength
 5,			// bDescriptorType
 0x81,			// bEndpointAddress	1 in
 0x03,			// bmAttributes		interrupt
 W(1<<EP1SSH),		// wMaxPacketSize	32
 1,			// bInterval
#endif
#if WUB
 9,			// bLength
 4,			// bDescriptorType
 HID,			// bInterfaceNumber
 0,			// bAlternateSetting
 2*!CDC,		// bNumEndpoints
 0xFF,			// bIfaceClass
 0,			// bIfaceSubClass
 0,			// bIfaceProtocol
 (4+HID)*(HID|CDC),	// iInterface
# if !CDC
// endpoint descriptor
 7,			// bLength
 5,			// bDescriptorType
 0x83,			// bEndpointAddress	3 in
 0x02,			// bmAttributes		bulk
 W(1<<EP3SSH),		// wMaxPacketSize	32
 0,			// bInterval
// endpoint descriptor
 7,			// bLength
 5,			// bDescriptorType
 0x04,			// bEndpointAddress	4 out
 0x02,			// bmAttributes		bulk
 W(1<<EP4SSH),		// wMaxPacketSize	16
 0,			// bInterval
# endif
#endif
#if CDC
# if IAD
// interface assocoation descriptor (required for Windows >7 after HID)
 8,			// bLength
 11,			// bDescriptorType	iface assoc
 HID+WUB,		// bFirstIface
 2,			// bIfaceCount
 2,			// bFuncClass		CDC
 2,			// bFuncSubClass
 1,			// bFuncProtocol
 4+HID+WUB,		// iFunction
# endif
// interface descriptor
 9,			// bLength
 4,			// bDescriptorType	iface
 HID+WUB,		// bInterfaceNumber
 0,			// bAlternateSetting
 1,			// bNumEndpoints
 2,			// bIfaceClass		CDC Control
 2,			// bIfaceSubClass	abstract control model
 1,			// bIfaceProtocol	V.25ter
 (4+HID+WUB)*(1-IAD),	// iInterface
//HEADER_FUNCTIONAL_DESCRIPTOR
 5,			// bFunctionLength
 0x24,			// bDescriptorType	CS_INTERFACE
 0,			// bDescriptorSubtype	header functional descriptor
 W(0x0110),		// bcdCDC		specification version, 1.1
//ABSTRACT_CONTROL_MANAGEMENT_FUNCTIONAL_DESCRIPTOR
 4,			// bFunctionLength
 0x24,			// bDescriptorType	CS_INTERFACE
 2,			// bDescriptorSubtype	abstract control management functional descriptor
 0x02,			// bmCapabilities
//UNION_FUNCTIONAL_DESCRIPTOR
 5,			// bFunctionLength
 0x24,			// bDescriptorType	CS_INTERFACE
 6,			// bDescriptorSubtype	union functional descriptor
 HID+WUB,		// bMasterInterface
 HID+WUB+1,		// bSlaveInterface0
//CALL_MANAGEMENT_FUNCTIONAL_DESCRIPTOR
 5,			// bFunctionLength
 0x24,			// bDescriptorType	CS_INTERFACE
 1,			// bDescriptorSubtype	call management functional descriptor
 3,			// bmCapabilities
 HID+WUB+1,		// bDataInterface
// endpoint descriptor
 7,			// bLength
 5,			// bDescriptorType
 0x82,			// bEndpointAddress	2 in
 0x03,			// bmAttributes		interrupt
 W(1<<EP2SSH),		// wMaxPacketSize	16
 8,			// bInterval
// interface descriptor
 9,			// bLength
 4,			// bDescriptorType
 HID+WUB+1,		// bInterfaceNumber
 0,			// bAlternateSetting
 2,			// bNumEndpoints
 10,			// bIfaceClass		CDC Data
 0,			// bIfaceSubClass
 0,			// bIfaceProtocol
 0,			// iInterface
// endpoint descriptor
 7,			// bLength
 5,			// bDescriptorType
 0x83,			// bEndpointAddress	3 in
 0x02,			// bmAttributes		bulk
 W(1<<EP3SSH),		// wMaxPacketSize	32
 0,			// bInterval
// endpoint descriptor
 7,			// bLength
 5,			// bDescriptorType
 0x04,			// bEndpointAddress	4 out
 0x02,			// bmAttributes		bulk
 W(1<<EP4SSH),		// wMaxPacketSize	16
 0,			// bInterval
#endif
};

#define STRDESC(name,str) static const PROGMEM struct{byte bLength,bDescType;int wString[sizeof(str)/2-1];}name={sizeof(str),3,str}

STRDESC(string0,L"\u0407");	// 0407 = deutsch, 0409 = englisch, Mehrfachauswahl möglich
STRDESC(string1,L"Tobias Lucas");
STRDESC(string2,L"Zwei Temperatur- und Feuchte-Sensoren");
STRDESC(string3,/*L"tobias.lucas@freenet.de?0001"*/L"heha@hrz.tu-chemnitz.de");

#if WUB
static const PROGMEM struct {
 unsigned a,b;
 unsigned long c;
 unsigned d,e,f,g,h,i,j,k,l,m,n;
 char o[16];
 unsigned p,q,r,s;
 int t[21];
 unsigned u;
 int v[40];
}MsOs20Desc = {
// Microsoft OS 2.0 descriptor set header (table 10)
 10,	// Descriptor size (10 bytes)
 0,	// MS OS 2.0 descriptor set header
 0x06030000,	// Windows version (8.1) (0x06030000)
 0xB2,	// Size, MS OS 2.0 descriptor set
// Microsoft OS 2.0 configuration subset header
 8,	// Descriptor size (8 bytes)
 1,	// MS OS 2.0 configuration subset header
 0,	// bConfigurationValue
 0xA8,	// Size, MS OS 2.0 configuration subset
// Microsoft OS 2.0 function subset header
 8,	// Descriptor size (8 bytes)
 2,	// MS OS 2.0 function subset header
 HID,	// interface number
 0xA0,	// Size, MS OS 2.0 function subset
// Microsoft OS 2.0 compatible ID descriptor (table 13)
 20,	// wLength
 3,	// MS_OS_20_FEATURE_COMPATIBLE_ID
 "WINUSB",
 4+21+1+40<<1,	//wLength: 
 4,	// wDescriptorType: MS_OS_20_FEATURE_REG_PROPERTY: 0x04 (Table 9)
 7,	// wPropertyDataType: REG_MULTI_SZ (Table 15)
 21<<1,	// wPropertyNameLength
 L"DeviceInterfaceGUIDs",	//bPropertyName
 40<<1,	// wPropertyDataLength
 L"{975F44D9-0D08-43FD-8B3E-127CA8AFFF9D}"	//bPropertyData
};

#define D(x) W(x),W((x)>>16)
static const PROGMEM byte BosDesc[] = {
 5,		// Length
 0x0F,		// Binary Object Store descriptor
 W(57),		// Total length
 2,		// Number of device capabilities
// WebUSB Platform Capability descriptor (bVendorCode == 0x01).
 0x18,		// Length
 0x10,		// Device Capability descriptor
 0x05,		// Platform Capability descriptor
 0x00,		// Reserved
 D(0x3408B638),W(0x09A9),W(0x47A0),0x8B,0xFD,0xA0,0x76,0x88,0x15,0xB6,0x65,  // WebUSB GUID
 W(0x0100),	// Version 1.0
 0x40,		// Vendor request code	(-> bRequest)
 0x01,		// Landing page		(-> wValueL)
// Microsoft OS 2.0 Platform Capability Descriptor (MS_VendorCode == 0x02)
 0x1C,		// Length
 0x10,		// Device Capability descriptor
 0x05,		// Platform Capability descriptor
 0x00,		// Reserved
 D(0xD8DD60DF),W(0x4589),W(0x4CC7),0x9C,0xD2,0x65,0x9D,0x9E,0x64,0x8A,0x9F,  // MS OS 2.0 GUID
 D(0x06030000),	// Windows version
 W(sizeof MsOs20Desc),	// Descriptor set length
 0x41,		// Vendor request code	(-> bRequest)
 0x00		// Alternate enumeration code
};
#undef D
//static_assert(sizeof BosDesc==57,"Wrong struct!");

#define LANDPAGE(name,scheme,str) static const PROGMEM struct{byte bLength,bDescType,bScheme;char string[sizeof(str)-1];}name={sizeof(str)+2,3,scheme,str}

LANDPAGE(LandPage,1,"www.tu-chemnitz.de/~heha/hs/bl/app.htm");
#endif

/**************************************************************************
 *  Variables - these are the only non-stack RAM usage
 **************************************************************************/

// zero when we are not configured, non-zero when enumerated
static volatile byte usb_configuration;

// the time remaining before we transmit any partially full
// packet, or send a zero-length packet.
static volatile byte flush_timer;

struct DCB{
// enum eStopbits{one,onehalf,two};
// enum eParity{none,odd,even,mark,space};
 unsigned long baudrate;
 byte stopbits,parity,databits;
}dcb = {9600,0,0,8};

/**************************************************************************
 *  Public Functions - these are the API intended for the user
 **************************************************************************/

// initialize USB
void usb_init(void) {
 HW_CONFIG();
 USB_FREEZE();			// enable USB
 PLL_CONFIG();			// config PLL
 while (!(PLLCSR & 1<<PLOCK));	// wait for PLL lock
 USB_CONFIG();			// start USB clock
 UDCON = 0;			// enable attach resistor
 usb_configuration = 0;
 UDIEN = 1<<EORSTE | 1<<SOFE;
 sei();
}

// return 0 if the USB is not configured,
// or the configuration number selected by the HOST
byte usb_configured(void) {return usb_configuration;}

// Transmit one byte. Return true on success, false on error
// Handles timeout based on SOF counter.
bool usb_putchar(char c) {
 static byte previous_timeout;
 byte timeout, intr_state;

	// if we're not online (enumerated and configured), error
 if (!usb_configuration) return false;
	// interrupts are disabled so these functions can be
	// used from the main program or interrupt context,
	// even both in the same program!
 intr_state = SREG;
 cli();
 UENUM = 3;
	// if we gave up due to timeout before, don't wait again
 if (previous_timeout) {
  if (!(UEINTX & 1<<RWAL)) {
   SREG = intr_state;
   return false;
  }
  previous_timeout = 0;
 }
	// wait for the FIFO to be ready to accept data
 timeout = UDFNUML + 4;
 for(;;) {
	// are we ready to transmit?
  if (UEINTX & 1<<RWAL) break;
  SREG = intr_state;
	// have we waited too long?
  if (UDFNUML == timeout) {
   previous_timeout = 1;
   return false;
  }
	// has the USB gone offline?
  if (!usb_configuration) return false;
	// get ready to try checking again
  intr_state = SREG;
  cli();
  UENUM = 3;
 }
	// actually write the byte into the FIFO
 UEDATX = c;
	// if this completed a packet, transmit it now!
 if (!(UEINTX & 1<<RWAL)) {
  UEINTX = 0x3A;
  flush_timer = 0;
 }else flush_timer = 2;
 SREG = intr_state;
 return true;
}

// immediately transmit any buffered output.
void usb_flush_output(void) {
 byte intr_state = SREG;
 cli();
 if (flush_timer) {
  UENUM = 3;
// HID kommt mit Short Packet nicht zurecht, daher mit Nullen auffüllen  
  while (UEINTX & 1<<RWAL) UEDATX = 0;
  UEINTX = 0x3A;	// Paket an SIE übergeben, ggf. rutscht sofort der zweite Puffer zur CPU
  flush_timer = 0;
 }
 SREG = intr_state;
}


// USB Device Interrupt - handle all device-level events.
// The transmit buffer flushing is triggered by the start of frame
ISR(USB_GEN_vect) {
 byte intbits = UDINT;
 UDINT = 0;
 if (intbits & 1<<EORSTI) {
  UENUM = 0;
  UECONX = 1;
  UECFG0X = 0;	// control endpoint
  UECFG1X = (EP0SSH-3)<<4 | 2/*SINGLE_BUFFER*/;
  UEIENX = 1<<RXSTPE;
  usb_configuration = 0;
 }
 if (intbits & 1<<SOFI && usb_configuration) {
  byte t = flush_timer;
  if (t && !(flush_timer = --t)) {
   UENUM = 3;
   while (UEINTX & 1<<RWAL) UEDATX = 0;
   UEINTX = 0x3A;
  }
 }
}


// Misc functions to wait for ready and send/receive packets
static inline void usb_wait_in_ready(void) {
 while (!(UEINTX & 1<<TXINI));
}
static inline void usb_send_in(void) {
 UEINTX = ~(1<<TXINI);
}
static inline void usb_wait_receive_out(void) {
 while (!(UEINTX & 1<<RXOUTI));
}
static inline void usb_ack_out(void) {
 UEINTX = ~(1<<RXOUTI);
}
static bool validEndpoint(byte ep) {
 return (HID && ep==0x81) || (CDC& (ep==0x82 || ep==0x83 || ep==0x04));
}

// Von-Neumannisiertes Speicherlesen (RAM oder Flash je nach Bit 15 der Adresse)
inline byte read_byte(const void*addr) {
 byte result;
 asm("	sbrc	%B1,7	\n"
 "	 lpm	%0,%a1	\n"
 "	sbrs	%B1,7	\n"
 "	 ld	%0,%a1	\n"
 :"=r"(result):"z"(addr));
 return result;
}

// Von-Neumannisierter Flash-Zeiger
inline const void*FP(const void*x) {return (void*)((unsigned)x+0x8000);}

// USB Endpoint Interrupt - endpoint 0 is handled here.  The
// other endpoints are manipulated by user-callable
// functions, and the start-of-frame interrupt.
ISR(USB_COM_vect) {
 UENUM = 0;
 byte intbits = UEINTX;
 if (intbits & 1<<RXSTPI) {
  byte bmRequestType = UEDATX,	// gesamtes SETUP-Paket in einzelne Register einlesen
	bRequest = UEDATX,
	wValueL = UEDATX,
	wValueH = UEDATX,
	wIndexL = UEDATX,
 UNUSED wIndexH = UEDATX,
	wLengthL = UEDATX,
	wLengthH = UEDATX,
	dlen;		// descriptor transfer length
  const void*addr;	// descriptor address (in FLASH space)
  UEINTX = ~(1<<RXSTPI | 1<<RXOUTI | 1<<TXINI);
  switch (bmRequestType) {
   case 0x80: switch (bRequest) {	// In,Std,Device
    case 6: switch (wValueH) {		// GetDescriptor((L=Index,H=DescType),LanguageID,Len)
     case 1: addr = FP(DeviceDesc);
gd1: dlen = read_byte(addr);
gd2: if (!wLengthH && dlen>wLengthL) dlen = wLengthL;
     byte n;	// needed outside loop for zero-length-packet detection!
     do{	// wait for host ready for IN packet
      byte i;
      do i = UEINTX; while (!(i & (1<<TXINI|1<<RXOUTI)));
      if (i & (1<<RXOUTI)) return;	// abort
      n = 1<<EP0SSH;			// send IN packet:
      if (n>dlen) n=dlen;		// it's a short packet, possibly zero!
      for (i = n; i; i--) UEDATX = read_byte(addr++);
      dlen -= n;
      usb_send_in();
     }while (dlen || n == 1<<EP0SSH);
     return;

     case 2: addr = FP(ConfigDesc); dlen = sizeof ConfigDesc; goto gd2;

     case 3: switch (wValueL) {
      case 0: addr = FP(&string0); goto gd1;
      case 1: addr = FP(&string1); goto gd1;
      case 2: addr = FP(&string2); goto gd1;
      case 3: addr = FP(&string3); goto gd1;
// Weitere String-Deskriptoren nur bei mehreren Interfaces
#if HID*(WUB|CDC)
      STRDESC(string4,L"HID-Interface");
      case 4: addr = FP(&string4); goto gd1;
#endif
#if WUB*(HID|CDC)
      STRDESC(string5,L"WebUSB");
      case 4+HID: addr = FP(&string5); goto gd1;
#endif
#if CDC
      STRDESC(string6,L"Serielle Schnittstelle");
      case 4+HID+WUB: addr = FP(&string6); goto gd1;
#endif
     }break;
#if WUB
     case 15: addr=FP(BosDesc); dlen=sizeof BosDesc; goto gd2;
#endif
#if HID
     case 0x21:
     case 0x22: goto usbtreeview;
#endif
    }break;

    case 8: {				// GetConfig(0,0,1)
     usb_wait_in_ready();
     UEDATX = usb_configuration;
     usb_send_in();
    }return;

    case 0: {				// GetStatus(0,0,2)
     UEDATX = 1;	// Bit0: Self-Power, Bit1: Remote Wakeup
     UEDATX = 0;
     usb_send_in();
    }return;
   }break;
   
   case 0x81: switch (bRequest) {	// In,Std,Iface
    case 0: {				// GetStatus(0,iface,2)
     UEDATX = 0;	// Stets 0
     UEDATX = 0;
     usb_send_in();
    }return;
    
#if HID
    case 6: switch (wIndexL) {		// GetDescriptor((L=?,H=?),iface,len)
     case 0: usbtreeview: switch (wValueH) {
      case 0x21:				// HidGetHidDesc(0x2100,iface,len)
      addr = FP(ConfigDesc)+HID_DESC_OFFSET; goto gd1;
      case 0x22: 		// HidGetReportDesc((L=descIndex,H=0x22),iface,len)
      addr = FP(HidRepDesc); dlen=sizeof HidRepDesc; goto gd2;
     }break;
    }break;
#endif

    case 10: {				// GetInterface(0,iface,1)
     UEDATX = 0;			// return one-and-only alternate setting 0
     usb_send_in();
    }return;
   }break;

   case 0x82: switch (bRequest) {	// In,Std,Endpoint
    case 0: {				// GetStatus(0,endpoint,2)
     if (!validEndpoint(wIndexL)) break;
     UENUM = wIndexL&7;
     byte answer = UECONX & 1<<STALLRQ ? 1 : 0;
     UENUM = 0;
     UEDATX = answer;	// Bit0: Endpoint stalled
     UEDATX = 0;
     usb_send_in();
    }return;
    case 18: {				// SyncFrame(0,endpoint,2)
     UEDATX = UDFNUML;
     UEDATX = UDFNUMH;
     usb_send_in();
    }break;
   }break;

   case 0x00: switch (bRequest) {	// Out,Std,Device
    case 1:				// ClearFeature(feature,0,0)
    case 3: return;			// SetFeature(feature,0,0)

    case 5: {				// SetAddr(addr,0,0)
     usb_send_in();
     usb_wait_in_ready();
     UDADDR = wValueL | 1<<ADDEN;	// Jetzt erst ändern!
    }return;

    case 9: {				// SetConfig(config,0,0)
     usb_configuration = wValueL;
     usb_send_in();
     for (byte i=0; i<4; i++) {
      UENUM = 1+i;
      UECONX = 1;
      static const byte cfg0[4] PROGMEM = {0xC1,0xC1,0x81,0x80};
      UECFG0X = pgm_read_byte(cfg0+i);	// Bulk/Interrupt In/Out
      static const byte cfg1[4] PROGMEM = {(EP1SSH-3)<<4|2,(EP2SSH-3)<<4|2,(EP3SSH-3)<<4|6,(EP4SSH-3)<<4|6};
      UECFG1X = pgm_read_byte(cfg1+i);	// size and single/double buffer
     }
     UERST = 0x1E;			// Togglebits rücksetzen
     UERST = 0;
    }return;
   }break;
   
   case 0x01: switch (bRequest) {	// Out,Std,Iface
    case 1:				// ClearFeature(sel,iface,0)
    case 3:				// SetFeature(sel,iface,0)
    case 11: return;			// SetInterface(alt_setting,iface,0)
   }break;

   case 0x02: switch (bRequest) {	// Out,Std,Endpoint
    case 1:				// ClearFeature(feature,endpoint,0)
    case 3: {				// SetFeature(feature,endpoint,0)
     if (wValueL) break;		// goto stall
     if (!validEndpoint(wIndexL)) break;
     usb_send_in();
     UENUM = wIndexL&=7;
     if (bRequest&2) UECONX = 1<<STALLRQ | 1<<EPEN;
     else{
      UECONX = 1<<STALLRQC | 1<<RSTDT | 1<<EPEN;
      UERST = 1<<wIndexL;		// Togglebit rücksetzen
      UERST = 0;
     }
    }return;
   }break;

   case 0x21: switch (wIndexL) {	// Out,Class,Iface
#if HID   
    case 0: switch (bRequest) {		// Iface=HID
     case 0x10: {			// HidSetReport((L=reportID,H=reportType),iface,len)
     }break;
    }break;
#endif
#if CDC
    case HID+WUB: switch (bRequest) {	// Iface=CDC
     case 0x20: usb_send_in(); return;	// CdcSetLineCoding(0,iface,7)
     case 0x22: usb_send_in(); return;	// CdcSetControlLineState(lineState,iface,0)
    }break;				// Bit 1 = RTS, Bit 0 = DTR
#endif
   }break;

#if HID|CDC   
   case 0xA1: switch (wIndexL) {	// In,Class,Iface
#if HID
    case 0: switch (bRequest) {		// Iface=HID
     case 1: {				// HidGetReport((L=reportID,H=reportType),iface,len)
      if (wIndexL) break;		// Interface 0 only
      dlen = wLengthH ? 255 : wLengthL;
      byte n;
      do{
       byte i;
     // wait for host ready for IN packet
       do i = UEINTX; while (!(i & (1<<TXINI | 1<<RXOUTI)));
       if (i & 1<<RXOUTI) return;	// abort
       n = 1<<EP0SSH;			// send IN packet
       if (n>dlen) n=dlen;
       for (i = n; i; i--) UEDATX = 0;	// of no data
       dlen -= n;
       usb_send_in();
      }while (dlen || n == 1<<EP0SSH);
     }return;
    }break;
#endif
#if CDC
    case HID+WUB: switch (bRequest) {
     case 0x20: {	// CdcGetSerialState(0,iface,2)
      UEDATX = 0;	// Bits: 6 = Overrun, 5 = Parity, 4 = Framing
      UEDATX = 0;	// 3 = RI, 2 = Break, 1 = DSR, 0 = DCD
      usb_send_in();
     }return;
     case 0x21: addr = &dcb; dlen = sizeof dcb; goto gd2;	// CdcGetLineCoding(0,iface,7)
    }break;
#endif
   }break;
#endif

#if WUB
   case 0xC0: switch (bRequest) {	// In, Vendor, Device
    case 0x40: addr=FP(&LandPage); goto gd1;	// wValue=1, wIndex=2
    case 0x41: addr=FP(&MsOs20Desc); dlen=sizeof MsOs20Desc; goto gd2;	// wValue=0, wIndex=7
   }break;
#endif

  }
 }
 UECONX = 1<<STALLRQ | 1<<EPEN;	// stall EP0 on switch-fallthrough or "break"
}
